//
// EregionAnalyse.cc
//
// Authors:  Richard Saxton (VEGA/ESA)
//           ESA XMM SOC
//
// Description: Task to calculate the optimum radius and centroid
//              for a point source extraction
//
//#include <ctime>
#include <iostream>
#include <fstream>

#include "Tracer.h"
#include "errstr.h"
#include "param.h"

#include "CalServer.h"
#include "Dal.h"

#include "EregionAnalyse.h"
#include "ImageRegion.h"
#include "CalSetter.h"
#include "STime.h"
#include "UpperLimitSolver.h"

/* Method: EregionAnalyse() - Constructor */
EregionAnalyse::EregionAnalyse():
srcreg(0),
bckreg(0),
_calset(0)
{
  PROC;

  /* Get an extension for temporary files based on time and pid */
    //srand(time(NULL) + getpid());    // sets the seed for the 
     srand(getpid());                  // random numbers by using the 
                                     // system time and the process id
    //cout<<"Random: "<<getpid()<<" "<<time(NULL)<<endl;

    _postfix = toString(rand());

    // _postfix = toString(STime().nowInMillisec());

  /* Get the parameters from this command line */
    string imagefile = stringParameter("imageset");
    string bckgndfile = stringParameter("bkgimageset");
    string exposurefile = stringParameter("exposuremap");
    string srcexp = stringParameter("srcexp");
    string backexp = stringParameter("backexp");
    double ulsig = realParameter("ulsig");

  /* Attempt to set the CAL state using the input image header */
    _calset = new CalSetter();
    _calset->fromImage(imagefile);
    _calset->setAccuracyLevel(EXTENDED);

  /* We can't currently optimise timing and burst mode data, so
     set status bad and exit early in these cases */
    Xmm::Mode windowMode = calServer.state()->mode()->get();

    if (windowMode == Xmm::FAST_TIMING ||
        windowMode == Xmm::FAST_BURST ||
        windowMode == Xmm::FAST_UNCOMPRESSED || 
        windowMode == Xmm::FAST_COMPRESSED) {
       
       setParameter("status",false);

       errstr<<" Timing/Burst mode - setting status false"
          << Sas::message(Msg::App, Msg::Verbose);

       ALTRETURN;
    }

  /* Set the instrument as a local variable */
    //Xmm::Instrument instr = calServer.instrument()->id();

  /* Set up an ImageRegion object for the source area */
    srcreg = new ImageRegion(imagefile, srcexp, 
                                "filtpixlist_src" + _postfix + ".ds:");

    errstr<<" Set up source image region"
          << Sas::message(Msg::App, Msg::Verbose);

  /* Check out a few methods */
    double delrad = srcreg->getRadius();
    double deltheta = srcreg->theta();
    double delcnts = srcreg->counts();
    double delflux = srcreg->flux();

    errstr<<" Radius: "<<delrad
          <<" Theta: "<<deltheta
          <<" Phi: "<<srcreg->phi() << endl
          <<" Counts: "<<delcnts << endl
          <<" numpix: "<<srcreg->number_of_pixels() << endl
          <<" flux: "<<delflux << endl
          <<" pixelArea: "<<srcreg->pixelArea()
          <<" xwid: "<<srcreg->xWidth()
          <<" ywid: "<<srcreg->yWidth()
          << Sas::message(Msg::App, Msg::Verbose);

  /* Process the background information: If a background map has been 
     supplied use this in conjunciton with the source spatial expression,
     If a map hasn´t been set and a background spatial expression has
     been supplied find the background flux per unit area from the source
     image. Otherwise use the value supplied in the parameter backval */

    bool backgnd_region_set=true;

    if (bckgndfile != "NotSet")
       bckreg = new ImageRegion(bckgndfile, srcexp, 
                                "filtpixlist_bgd" + _postfix + ".ds:");
    else if (backexp != "NotSet")
    {
       bckreg = new ImageRegion(imagefile, backexp, 
                                "filtpixlist_bgd" + _postfix + ".ds:");
    } 
    else {
       bckreg=0;
       backgnd_region_set=false;
    }

   /* Find the background per unit area */
    double bck_flux=0;
    if (backgnd_region_set) {

       bck_flux = bckreg->flux();

       errstr<<" Backgnd: Radius: "<<bckreg->getRadius()
          <<" Theta: "<<bckreg->theta()
          <<" Counts: "<<bckreg->counts()<< endl
          <<" numpix: "<<bckreg->number_of_pixels() << endl
          <<" pixelArea: "<<bckreg->pixelArea()
          <<" xwid: "<<bckreg->xWidth()
          <<" ywid: "<<bckreg->yWidth()
          <<" Background flux: "<<bck_flux
          << Sas::message(Msg::App, Msg::Verbose);

    } else {  // get backgnd flux from command line
       bck_flux = realParameter("backval");
    }

  /* Centroid the source */
    CoordPair centre = srcreg->centroid(5);

  /* Find the centroid in detector coords */
    //CoordPair detCentre = srcreg->conv2Det(centre);

    //cout<<"Det centroid: "<<detCentre.x()<<" "<< detCentre.y()<<endl;

  /* Find the number of counts in the source area */
    double src_cnts = srcreg->counts();

  /* Should get energy from selection expression but lets approximate
     to 1.5 keV for now */
    double energy=1500.;

  /* Get the off-axis angle and azimuthal angle of the source box */
    double theta = srcreg->theta();
    double phi = srcreg->phi();
    double src_rad = srcreg->getRadius();

  /* Find the optimum radius (in arcsecs) of the source box and
     convert to image units  - using the EXTENDED (circular) PSF model */
    _calset->setAccuracyLevel(EXTENDED);
    double optrad = optimumRadius(src_cnts, src_rad, bck_flux,
                                  energy, theta, phi);
    double optrad_image = optrad * srcreg->asecToImgUnits();

  /* Find the optimum radius (in arcsecs) of the source box and
     convert to image units  - using the ELLBETA (elliptical) PSF model */
    FEllipse * optellip = optimumEllipse(src_cnts, src_rad, bck_flux,
                              energy, theta, phi, centre, 
                              srcreg->calcNormal(phi));

  /* Find the exposure time for the source. If an exposure map has been
     provided use this, otherwise just use the exposure keyword from 
     the source image which is only correct if the source is on-axis */
    double exposure = 0.0;
    if (exposurefile != "NotSet") {

      // If REF keywords are missing from the exposure file header
      // Make a copy of the exposure file, add in the keywords and use this
       string tempexpfile = expfile_copy(exposurefile, imagefile); 

       expreg = new ImageRegion(tempexpfile, srcexp, 
                                "filtpixlist_exp" + _postfix + ".ds:");
       exposure = expreg->centralPixelValue();

    } else {
       exposure = srcreg->exposure();
    }

  /* Calculate the background subtracted, EEF corrected, source count rate
     and upper limit */
    // double cl=0.954;  ! Set 2-sigma upper limit
    calcSrcCountRate(srcreg, bckreg, bck_flux, backgnd_region_set,
                                  energy, exposure, ulsig);

  /* Get the initial source region centre in RA/DEC */
    CoordPair initcentre = srcreg->celestialCentre();
    double x_coord = initcentre.x();
    double y_coord = initcentre.y();

   // Or use detector coordinates if image is in detector coords 
    if (srcreg->imcoordType() == "DET") {
       x_coord = srcreg->centre().x();
       y_coord = srcreg->centre().y();
    }
    
    /* 
       AI:SOC-SPR-2687. 
       For some reason in Mac OS X you cannot call a method inside a 
       errstr statement. So we have defined some variables where we store the 
       values and the we print them out using the errstr and the SAS log infrastucutre.
     */

    double cnts_per_pixel = srcreg->countsperpix();
    double x_centre=centre.x();
    double y_centre=centre.y();
    double encircle_fac=_source.ee;
    double bkg_sub_count = _source.bkgsub_counts;
    double bkg_sub_err =_source.bkgsub_counts_error;
    double bkg_sub_cr = _source.bkgsub_count_rate;
    double bkg_sub_cr_err = _source.bkgsub_count_rate_error;
    double upplim=_source.upplim;
   
   // Elliptical params
    double xrad = optellip->radii()[0];
    double xrad_image = xrad * srcreg->asecToImgUnits();
    double yrad = optellip->radii()[1];
    double yrad_image = yrad * srcreg->asecToImgUnits();
    double rotang = optellip->rotAngles()[0];

   // Find the central CCD for the source position
   // Ccd ccd_cent = srcreg->centralCcd();
   // vector<Ccd> ccdlist = srcreg->ccdList();

    //string ccds = "";
    //for (unsigned int i=1 ; i<ccdlist.size(); i++)
      //  if (ccdlist[i] != 0) ccds += " " + toString(i);

  /* Output the selection strings for the optical regions */
   // First set the coord type of the output string
    string unitbit = "";
    if (srcreg->coordType() == "DET") {unitbit = "DET";}

    string scircle = "(" + unitbit + "X," + unitbit + "Y) in CIRCLE("  
                   + toString(x_centre) + "," 
                   + toString(y_centre) + "," + toString(optrad_image) + ")";

    string sellipse = "(" + unitbit + "X," + unitbit + "Y) in ELLIPSE(" 
                    + toString(x_centre) + "," + toString(y_centre) + "," 
                    + toString(xrad_image) + "," 
                    + toString(yrad_image) + "," + toString(rotang) + ")";    

  /* Output results to a file or screen */

    ostringstream outs;
    outs.precision(8);

       //outfile<<" input region centre: "<<x_coord <<" "<<y_coord <<endl
    outs<<" input region centre: "<<x_coord <<" "<<y_coord <<endl
        <<" counts in source region: "<<src_cnts <<endl
        <<" src region cnts per pixel: " <<cnts_per_pixel <<endl
        <<" exposure time: "<<exposure<<endl
        <<" xcentroid: "<<x_centre <<endl
        <<" ycentroid: "<< y_centre <<endl
        <<" optradius: "<<optrad << " arcsecs "<<optrad_image << " image units" <<endl
        <<" optellip: X radius: "<<xrad<<" arcsecs "<<xrad_image << " image units Y radius: " <<yrad<<" arcsecs "<<yrad_image << " image units rotangle: "<<rotang<<endl
        <<" encircled energy factor: "<<  encircle_fac <<endl
        <<" Bckgnd subtracted source cnts: "<< bkg_sub_count <<" +/- "<< bkg_sub_err <<endl
        <<" Bckgnd subtracted source c/r: "<< bkg_sub_cr <<" +/- "<< bkg_sub_cr_err <<endl
        <<" Statistical upper limit c/r: "<<upplim<<" c/s" << endl <<" "<<endl
        <<" SASCIRCLE: "<<scircle<<endl
        <<" SASELLIPSE: "<<sellipse<<endl;

    if (booleanParameter("withoutputfile")) {

      /* Open a stream to an output text file */
       ofstream outfile;
       string textfile=stringParameter("output");
       outfile.open(textfile.c_str());

       outfile<<outs.str();

       outfile.close();
    
    } else { // Write to standard error

       errstr<<outs.str()
             <<Sas::message(Msg::App, Msg::Always);
    }

  /* Set output parameters */
    setParameter("xcentroid",centre.x());
    setParameter("ycentroid",centre.y());
    setParameter("optradius",optrad_image);
    setParameter("optellipxrad",xrad_image);
    setParameter("optellipyrad",yrad_image);
    setParameter("optelliprot",rotang);
    setParameter("srccnts",src_cnts);
    setParameter("status",true);

  ENDPROC;
}

/* Destructor */
EregionAnalyse:: ~EregionAnalyse()
{
  PROC;

    delete srcreg;            srcreg=0;
    delete bckreg;            bckreg=0;
    delete _calset;           _calset=0;

  ENDPROC;
};

/* Method: optimumRadius - finds the optimum radius for source extraction
        from the source and background counts */
double EregionAnalyse::optimumRadius(double src_cnts, double sboxrad, 
                     double bck_flux,
                     double energy, double theta, 
                     double phi)
{
  /* Set up the encircled energy function */
   PsfDataServer * psfs = 0; psfs = calServer.getAtom(psfs);
   Psf & p = psfs->psf(energy, theta, phi);
   _psf = &p;

   errstr<<"EEF set up with Energy: "<<energy<<" Theta: "<<theta
         <<" Phi: "<<phi
            << Sas::message(Msg::App, Msg::Verbose);

  /* Get the total number of counts in source */
   /* Find the fraction of counts in the source box radius */
   double asec2mm = 27.5;  // This is cack should get it from PSF file
                           // but not going to change...
   double test_rad = sboxrad / asec2mm;
   double ee = _psf->encircledEnergy(test_rad); 

   /* Find the background counts in the source box */
   double bck_cnts = bck_flux * M_PI * sboxrad * sboxrad;

   errstr<<" Back: "<<bck_cnts<<" "<<bck_flux<<" "<<sboxrad
         <<" EE(2.4) "<<_psf->encircledEnergy(2.4)
            << Sas::message(Msg::App, Msg::Verbose);


   double totsrc = (src_cnts - bck_cnts) / ee;

   errstr<<" SRC: "<<totsrc<<" "<<src_cnts<<" "<<bck_cnts
            << Sas::message(Msg::App, Msg::Verbose);

   double snr=0, snrmax=0, maxrad=0;

  /* Loop over test radii between 0.0 and 5.0 arcminutes in one arcsec bins */

   for (int i=1; i<301; i++)
   {
    /* Get the fraction of source counts within a circle of this radius.
       NB: This takes forever to complete unless the PSF mode is EXTENDED
       or ELLBETA so use a mickey-mouse formulation for other modes */

      if (calServer.state()->accuracyLevel()->get() == EXTENDED ||
          calServer.state()->accuracyLevel()->get() == EXTENDED) {
          // calServer.state()->accuracyLevel()->get() == ELLBETA) {
         test_rad = i / asec2mm;  // test radius is in mm (CamCoords2) 
         ee = _psf->encircledEnergy(test_rad);
       }
      else {
         ee = i / 300.0;
       }

    /* Calculate the S/N */
      snr = totsrc * ee / sqrt (totsrc * ee + bck_flux * M_PI
                                   * i * i);
      errstr<<" Optimum try: "<<i<<" "<<totsrc<<" "<<ee<<" "
            <<bck_flux<<" "<<M_TWO_PI<<" "<<snr
            << Sas::message(Msg::App, Msg::Noisy);

    /* Update the maximum */
      if (snr > snrmax) 
      {
         snrmax = snr;
         maxrad = i;
      }
   }

  /* Set artifical mimimum extent if optimum radius too small */
   maxrad = maxrad < 10 ? 10 : maxrad;
 
   errstr<<" Optimum radius is: "<<maxrad
   << Sas::message(Msg::App, Msg::Verbose);

   RETURN maxrad;
}

/* Method: optimumEllipse - finds the optimum ellipse for source extraction
        from the source and background counts and source position */
FEllipse * EregionAnalyse::optimumEllipse(double src_cnts, 
                     double src_rad, 
                     double bck_flux,
                     double energy, double theta, 
                     double phi, CoordPair centre, double orient)
{
  /* Set up the PSF for the elliptical model. For the moment supress
     messages due to missing ELLBETA model */
   _calset->setAccuracyLevel(ELLBETA);
   PsfDataServer * psfs = 0; psfs = calServer.getAtom(psfs);
   Psf & p = psfs->psf(energy, theta, phi);

  /* The PSF is calibrated for an ellipse which has an ellipticity that
     depends on off-axis angle and photon energy. Get the ellipticity */
   double ellipticity = p.ellipticity();
   //double ellipticity = 0.0;
   
  /* Find the radius of the major-axis of the optimum ellipse */
   double optrad = optimumRadius(src_cnts, src_rad, bck_flux,
                                  energy, theta, phi);

  /* Set the major and minor axes */
   double xrad = optrad;
   double yrad = optrad * sqrt(1.0 - ellipticity*ellipticity);

   //cout<<"Ellipticity: "<<ellipticity<<endl;

  /* Construct the extraction expression for the optimum ellipse */
   return new FEllipse( FPoint(centre.x(), centre.y()), 
                                    xrad, yrad, orient, true);
}

/* Method: calcSrcCountRate - returns the background subtracted source
                               count rate * EEF */
void EregionAnalyse::calcSrcCountRate(ImageRegion * srcreg,
                     ImageRegion * bckreg,
                     double bck_flux,
                     bool backgnd_region_set,
                     double energy, double exposure, double cl) 
{
  /* Set up the encircled energy function */
   _calset->setAccuracyLevel(EXTENDED);
   PsfDataServer * psfs = 0; psfs = calServer.getAtom(psfs);
   Psf & p = psfs->psf(energy, srcreg->theta(), srcreg->phi());
   _psf = &p;

  /* Get the total number of counts in source */
   /* Find the fraction of counts in the source box radius */
   double asec2mm = 27.5;
   double test_rad = srcreg->getRadius() / asec2mm;
   double ee = _psf->encircledEnergy(test_rad); 

   double src_bck_cnts=0, totsrc=0, toterror=0;

   /* Find the background counts in the source box  and then the
      background subtracted source count rate and error */
   if (backgnd_region_set)
   {
      double area_ratio = srcreg->area() / bckreg->area();
      src_bck_cnts = bckreg->counts() * area_ratio;

      errstr<<" Back: "<<src_bck_cnts<<" "<<area_ratio
         <<" EE(2.4) "<<_psf->encircledEnergy(2.4)
            << Sas::message(Msg::App, Msg::Verbose);

     /* Calc bckgnd subtracted counts corrected for PSF */
      totsrc = (srcreg->counts() - src_bck_cnts) / ee;
      toterror = sqrt(srcreg->counts() + bckreg->counts() 
                                       * area_ratio*area_ratio) / ee;

   } else { // Using entered bckgnd flux
      src_bck_cnts = bck_flux * srcreg->area();
      totsrc = (srcreg->counts() - src_bck_cnts) / ee;
      // toterror = 0.0;  // set to zero in this case
      toterror = sqrt(srcreg->counts()) / ee;
   }

 /* Set the background subtracted source count rate and error */
   _source.bkgsub_counts = totsrc;
   _source.bkgsub_counts_error = toterror;
   _source.ee = ee;
   
   if (exposure > 0.0) {
     _source.bkgsub_count_rate = totsrc / exposure;
     _source.bkgsub_count_rate_error = toterror / exposure;
   } else {
     _source.bkgsub_count_rate = 0.0;
     _source.bkgsub_count_rate_error = 0.0;
   }

  // Calculate the upper limit. For lowish counts use the Bayesian
  // statistics. NB: This isn't robust for N>80.
   _source.upplim = 0.0;

   if (exposure > 0.0) {  // Leave upper limit as 0 if exposure <= 0
     if (srcreg->counts() < 80) { // Use Bayesian statistics
        _source.upplim = calcUpplim(srcreg->counts(), src_bck_cnts, cl) 
                                                                / exposure;
     } else {// Use source counts + 2 x error
      
    /* Need to convert between confidence interval and sigma.
       Can't find the equation right now - so use bodge */
       double fac=0.0;
       if ( cl>0.67 && cl<0.69)
          fac=1.0;
       else if ( cl>0.98)
          fac=3.0;
       else
          fac=2.0;
     
       _source.upplim = ( (srcreg->counts() - src_bck_cnts) +
                     fac * ( toterror * ee ) ) / exposure;
     }
   }
}

/* Method: expfile_copy - makes a copy of exposure file and adds in REF keys
                          if necessary */
const string EregionAnalyse::expfile_copy(const string & exposurefile,
										  const string & imagefile)
{
   // Set the name of a temporary copy of the exposure file
     string tempexpfile = "temp_expfile_copy"+_postfix+".ds";

     DataSet * tempexpset = dataSetServer->clone(
                      exposurefile,
                      tempexpfile,
                      //Dal::Modify);
                      Dal::Temp);

   /* Check if all the necessary REF keywords are present. These 
      keywords are used to convert celestial positions to X/Y */
     if (! tempexpset->hasAttribute("REFXCRPX"))
     {

      // Copy over REF keywords from the image file
        DataSet * imageset = dataSetServer->open(imagefile, DataSet::Read); 

        if (imageset->hasAttribute("REFXCRPX"))
        {
         * tempexpset->addAttribute("REFXCRPX") = 
                    imageset->attribute("REFXCRPX")->asInt();
         * tempexpset->addAttribute("REFXCDLT") = 
                    imageset->attribute("REFXCDLT")->asReal();
         * tempexpset->addAttribute("REFYCDLT") = 
                    imageset->attribute("REFYCDLT")->asReal();
         * tempexpset->addAttribute("REFYCRPX") = 
                    imageset->attribute("REFYCRPX")->asInt();
         * tempexpset->addAttribute("REFXCTYP") = 
                    imageset->attribute("REFXCTYP")->asString();
         * tempexpset->addAttribute("REFYCTYP") = 
                    imageset->attribute("REFYCTYP")->asString();
         * tempexpset->addAttribute("REFXCRVL") = 
                    imageset->attribute("REFXCRVL")->asReal();
         * tempexpset->addAttribute("REFYCRVL") = 
                    imageset->attribute("REFYCRVL")->asReal();
        }
        dataSetServer->close(imageset);
     }

   /* Now check whether the CTYPE1 keywords are in the right place */
     Array * outexparray = tempexpset->array(tempexpfile+":");

     if (! outexparray->hasAttribute("CTYPE1")) 
     {
        /* Copy in keywords from the main header to the array header */
         * outexparray->addAttribute("CTYPE1") =
                    tempexpset->attribute("CTYPE1")->asString();
         * outexparray->addAttribute("CTYPE2") =
                    tempexpset->attribute("CTYPE2")->asString();
         * outexparray->addAttribute("CRPIX1") =
                    tempexpset->attribute("CRPIX1")->asString();
         * outexparray->addAttribute("CRVAL1") =
                    tempexpset->attribute("CRVAL1")->asString();
         * outexparray->addAttribute("CDELT1") =
                    tempexpset->attribute("CDELT1")->asString();
         * outexparray->addAttribute("CRPIX2") =
                    tempexpset->attribute("CRPIX2")->asString();
         * outexparray->addAttribute("CRVAL2") =
                    tempexpset->attribute("CRVAL2")->asString();
         * outexparray->addAttribute("CDELT2") =
                    tempexpset->attribute("CDELT2")->asString();
     }

     //dataSetServer->close(tempexpset);

     RETURN tempexpfile;
}
/* EregionAnalyse::calcUpplim - calculates the upper limit for total counts
,n, background counts, b and confidence limit, cl */
double EregionAnalyse::calcUpplim(double n, double b, double cl)
{
   UpperLimitSolver u = UpperLimitSolver(static_cast<long int>(n),b);
   std::pair<double, double> limits = u.bayesian(cl);
   //double lower = limits.first, upper=limits.second;
   double upper=limits.second;

   //cout<<"LIMITS: "<<lower<<" "<<upper<<endl;

   RETURN upper;
}

//const string EregionAnalyse::getPostFix()
//{
 //  struct timeb stv;
  // ftime(&stv);
   
 //  RETURN toString(stv.time*1000 + stv.millitm);
//}
