//
// SpectrumMaker.cc
//
// Authors:  Richard Saxton (VEGA/ESA)
//           ESA XMM SOC
//
// Description: Creates a spectrum using evselect
//
#include <Tracer.h>
#include <metatask.h>
#include <errstr.h>
#include <param.h>
#include <Dal.h>
#include <evselect.h>

#include "SpectrumMaker.h"

SpectrumMaker::SpectrumMaker(string spectrumfile, 
                     DatafileAttributes * specAttributes, string expression) : 
_spectrumfile(spectrumfile)
{
  PROC;

   /* Set up the parameters for calling evselect */
    openParameters("evselect");

   /* Ideally the selection expression should contain just a source
      spatial region. Everything else is defaulted here */

   /* Set the variable parameters */
    setParameter("spectrumset",_spectrumfile);
    setParameter("table",specAttributes->datafile());

    /* Choose the spectral range, pattern range and flag settings
                                      depending on the instrument */
    switch (specAttributes->instrumentID())
    {
    case Xmm::EMOS1 :
    case Xmm::EMOS2 :
      {
        _specbin=15;
        _specmin=0.0;
        _specmax=11999.0;
        _pattexp="&&(PATTERN<=12)";
        _flagexp="&&(#XMMEA_EM)";
        break;
      }
    case Xmm::EPN :
      {
        _specbin=5;
        _specmin=0.0;
        _specmax=20479.0;
        _pattexp="&&(PATTERN<=4)";
        _flagexp="&&(FLAG==0)";
        break;
      }
    default:
      errstr << "Dont recognise the instrument "<<specAttributes->instrument()
             << Sas::error("invalidInstrument");
    }

    /* Now set the spectral parameters */
    setParameter("spectralbinsize",_specbin);
    setParameter("specchannelmin",_specmin);

errstr<<"Setting channel min to 0"
<< Sas::message(Msg::App, Msg::Verbose);

    setParameter("specchannelmax",_specmax);

errstr<<"Set channel max to "<<_specmax
<< Sas::message(Msg::App, Msg::Verbose);

    /* Add the flag and pattern selection into the evselect expression. 
       If a pattern selection already exists then this will not have 
       any effect with current settings */ 

    setParameter("expression",expression+_pattexp+_flagexp);
       
  /* Set the other parameters */
    setParameter("withfilteredset",false);
    setParameter("keepfilteroutput",false);
    setParameter("flagcolumn","EVFLAG");
    setParameter("flagbit",-1);
    setParameter("filtertype","expression");
    setParameter("writedss",true);          
    setParameter("cleandss",false);          
    setParameter("updateexposure",true);          
    setParameter("filterexposure",true);          
    setParameter("energycolumn","PI");
    setParameter("withzcolumn",false);
    setParameter("withspectrumset",true);
    setParameter("withspecranges",true);
    setParameter("withrateset",false);
    setParameter("withtimeranges",false);
    setParameter("withhistogramset",false);

  /* Now run evselect using these parameters */ 
    task(evselect);
   
  ENDPROC;
}

/* Method to add standard keywords to header */
void SpectrumMaker::addkeywords(string bgdfile,
                                string rspfile, string arffile)
{
 PROC;

   errstr
    << "Entered addkeywords"
    << Sas::message(Msg::App, Msg::Verbose);

  /* Add keywords into the SPECTRUM extension of the file */
   DataSet * spectrumset = dataSetServer->open( _spectrumfile, 
                         DataSet::Modify );

   errstr
    << "Getting the table"
    << Sas::message(Msg::App, Msg::Verbose);

   Table * spectrumtable = spectrumset->table("SPECTRUM", Dal::Modify);

   errstr
    << "Adding keywords into table: " << spectrumtable 
    << Sas::message(Msg::App, Msg::Verbose);

   * spectrumtable->addAttribute("BACKFILE", "Name of background file") = bgdfile;
   errstr
    << "Adding keywords into table: 1 " << bgdfile
    << Sas::message(Msg::App, Msg::Verbose);

   * spectrumtable->addAttribute("RESPFILE", "Name of response file") = rspfile;
   * spectrumtable->addAttribute("ANCRFILE", "Name of ARF file") = arffile;

   errstr
    << "Adding keywords into table: 2 " << rspfile
    << Sas::message(Msg::App, Msg::Verbose);

   dataSetServer->close(spectrumset);

 ENDPROC;
}

/* Destructor */
SpectrumMaker::~SpectrumMaker()
{
  PROC;
  ENDPROC;
}


