//
// ecoordconv.cc
//
// Authors:  Richard Saxton (VEGA/ESA)
//           ESA XMM SOC
//
// Description: Task to convert a point from any to all systems
//
//#include <ctime>
#include "Tracer.h"
#include "errstr.h"
#include "param.h"

#include "CalServer.h"
#include "Dal.h"

#include "ImageRegion.h"
#include "CalSetter.h"
#include "STime.h"
#include "ecoordconv.h"

void ecoordconv()
{
  PROC;

  /* Get an extension for temporary files based on time and pid */
    //srand(time(NULL) + getpid());    // sets the seed for the 
    srand(getpid());                  // random numbers by using the 
                                     // system time and the process id

    string _postfix = toString(rand());

    // _postfix = toString(STime().nowInMillisec());

  /* Get the parameters from this command line */
    string imagefile = stringParameter("imageset");
    string srcexp = stringParameter("srcexp");

  /* If entered as separate position parameters construct a srcexp string */
    if  (booleanParameter("withcoords")) {
        double x = realParameter("x");
        double y = realParameter("y");
        string coordtype = stringParameter("coordtype");
        
        string cstring="";
        string crad="100.0";

        if (uppercase(coordtype) == "DET") 
             {cstring = "(DETX,DETY)";}
        else if (uppercase(coordtype) == "POS") 
             {cstring = "(X,Y)";}
        else if (uppercase(coordtype) == "EQPOS") 
             {cstring = "(RA,DEC)"; crad="0.01";}
        else {
            errstr << "Coordtype unknown: "<<coordtype
             << Sas::error("invalidCoordType");
        }
        srcexp = cstring + " in circle(" + toString(x) + "," 
                                         + toString(y) + "," + crad + ")";
    }

  /* Attempt to set the CAL state using the input image header */
    CalSetter * _calset = new CalSetter();
    _calset->fromImage(imagefile);
    _calset->setAccuracyLevel(EXTENDED);

  /* We can't currently optimise timing and burst mode data, so
     set status bad and exit early in these cases */
    Xmm::Mode windowMode = calServer.state()->mode()->get();

    if (windowMode == Xmm::FAST_TIMING ||
        windowMode == Xmm::FAST_BURST ||
        windowMode == Xmm::FAST_UNCOMPRESSED || 
        windowMode == Xmm::FAST_COMPRESSED) {
       
  //     setParameter("status",false);

       errstr<<" Timing/Burst mode data is not supported"
             << Sas::error("UnsupportedMode");

       ALTRETURN;
    }

  /* Set up an ImageRegion object for the source area */
    ImageRegion * srcreg = new ImageRegion(imagefile, srcexp, 
                                "filtpixlist_src" + _postfix + ".ds:");

    errstr<<" Set up source image region"
          << Sas::message(Msg::App, Msg::Verbose);

  /* Find the central CCD for the source position and otehr CCDs*/
    Ccd ccd_cent = srcreg->centralCcd();
    vector<Ccd> ccdlist = srcreg->ccdList();

    string ccds = "";
    for (unsigned int i=1 ; i<ccdlist.size(); i++)
        if (ccdlist[i] != 0) ccds += " " + toString(i);

  /* Set a few local variables */
    double theta  = srcreg->theta();
    double phi = srcreg->phi();

    CoordPair radec = srcreg->celestialCentre();
    double ra = radec.x();
    double dec = radec.y();

    CoordPair detcen = srcreg->conv2Det(srcreg->centre());
    double detx = detcen.x();
    double dety = detcen.y();

    CoordPair rawcen = srcreg->conv2Raw(srcreg->centre());
    double rawx = rawcen.x();
    double rawy = rawcen.y();

    CoordPair poscen = srcreg->conv2Pos(srcreg->centre());
    double posx = poscen.x();
    double posy = poscen.y();

  /* Output the result */
    errstr<<" Region Centre: "<<endl
          <<" Theta: Phi: "<<theta<<" "<<phi<<endl
          <<" X: Y: "<<posx<<" "<<posy << endl
          <<" DETX: DETY: "<<detx<<" "<<dety << endl
          <<" RA: DEC: "<<ra<<" "<<dec << endl
          <<" RAWX: RAWY: "<<rawx<<" "<<rawy << endl
	  <<" CCD(s): "<<ccds<<" centred on CCD: "<<ccd_cent<<endl
          << Sas::message(Msg::App, Msg::Always);

  ENDPROC;
}
