//
// especget.cc
//
// Authors:  Richard Saxton (VEGA/ESA)
//           ESA XMM SOC
//
// Description: Metatask which runs evselect, arfgen and rmfgen to 
//              produce source and background spectra and the appropriate
//              ARF and RMF.
//
//#include <set>          // C++
//#include <algorithm>
//#include <strstream.h>
//#include <unistd.h>
#include <Tracer.h>
#include <metatask.h>
#include <errstr.h>
#include <param.h>
#include "Evaluator.h"

// SAS packages
#include "arfgen.h"   // tasks
#include "DatafileAttributes.h"   // tasks

#include "especget.h"
#include "SpectrumMaker.h"
#include "ArfMaker.h"
#include "RmfMaker.h"
#include "BackscaleMaker.h"
#include "FindBestRmfName.h"

void especget()
{
  PROC;

  /* Set the source, backgnd and arf filenames */
    string sourceset, backset, arfset, outrmfset;    

    if  (booleanParameter("withfilestem")) {
        string filestem = stringParameter("filestem");
        sourceset = filestem+"_src.ds";
        backset   = filestem+"_bgd.ds";
        arfset    = filestem+"_src.arf";
        outrmfset = filestem+"_src.rmf";
    } else {
        sourceset = stringParameter("srcspecset");
        backset   = stringParameter("bckspecset");
        arfset    = stringParameter("srcarfset");
        outrmfset = stringParameter("srcrmfset");
    }

  /* Get the parameters from this command line */
    string srcexp = stringParameter("srcexp");
    string backexp = stringParameter("backexp");
    bool withbadpixcorr = booleanParameter("withbadpixcorr");
    bool useodfatt = booleanParameter("useodfatt");
    string eventfile = stringParameter("table");
    bool withrmfset = booleanParameter("withrmfset");
    string rmfset = stringParameter("rmfset");
    bool extendedsource = booleanParameter("extendedsource");
    bool witharfset = booleanParameter("witharfset");
    unsigned xbins = (unsigned) intParameter("detxbins");
    unsigned ybins = (unsigned) intParameter("detybins");

  /* Check that the input expressions are valid */
    if (! Evaluator::isValidExpression(srcexp)) {
       errstr<<"Invalid expression: "<<srcexp
             << Sas::error("InvalidExpression");
    }
    else if (! Evaluator::isValidExpression(backexp)) {
       errstr<<"Invalid expression: "<<backexp
             << Sas::error("InvalidExpression");
    }

  /* Read observation attributes from the event file */
    DatafileAttributes * specAttributes = new DatafileAttributes(eventfile);

  /* Create the source spectrum */
    errstr<<"Creating source spectrum..."
          << Sas::message(Msg::App, Msg::Sparse);

    SpectrumMaker sspec = SpectrumMaker(sourceset, specAttributes,
                                             srcexp);

  /* Create the background spectrum */
    errstr<<"Creating background spectrum..."
          << Sas::message(Msg::App, Msg::Sparse);

    SpectrumMaker bspec = SpectrumMaker(backset, specAttributes,
                                             backexp);

  /* Set the source area keyword (BACKSCAL) for the spectrum and
     create the ARF */
    if (witharfset) 
    {
    errstr<<"Creating the ARF and calculating area for the source..."
          << Sas::message(Msg::App, Msg::Sparse);

    ArfMaker::ArfMaker(sourceset,
                       arfset,
                       withbadpixcorr,
                       useodfatt,
                       eventfile,
                       withrmfset,
                       rmfset,
                       extendedsource,
                       xbins, ybins,
                       true);

  /* Set the background area keyword */
    errstr<<"Calculating area of the background region..."
          << Sas::message(Msg::App, Msg::Sparse);

  /* Make sure that the temporary file doesn't exist to avoid problems
     with the dataset not being closed down correctly */
    string temparf =  "temparf.arf";
    if ( dataSetServer->exists(temparf))
          dataSetServer->remove(temparf);

    ArfMaker::ArfMaker(backset,
                       temparf,
                       withbadpixcorr,
                       useodfatt,
                       eventfile,
                       withrmfset,
                       rmfset,
                       extendedsource,
                       xbins, ybins,
                       false);
    }

  /* Create the RMF if requested. Otherwise set to the given name*/
    if (! withrmfset) 
    {
       errstr<<"Creating the RMF..."
          << Sas::message(Msg::App, Msg::Sparse);

       RmfMaker::RmfMaker(sourceset,outrmfset,extendedsource);

       rmfset = outrmfset;
    }
    else 
    {          // If the rmfset name has not been given find the
               // best default
       if ( rmfset == " " ) {
          rmfset = FindBestRmfName::getName(sourceset);
       }
    }

  /* Set the back, arf and rmf filenames in the source spectrum header */
    sspec.addkeywords(backset, rmfset, arfset);

  /* Shutdown */
    closeParameters();

  ENDPROC;
}

